
# This file was generated by 'versioneer.py' (0.14) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

version_version = '0.13.0'
version_full = 'e924ffa17a010bf6648b54adc6f8d4b4b7ec8755'
def get_versions(default={}, verbose=False):
    return {'version': version_version, 'full': version_full}

